{
    Файл является частью библиотеки RubiRobotLib

    © 2017-2021 Слинкин Д.А.

    Обработка кнопок контроллера Lego EV3

    Подробная информация о лицензировании находится в файле README,
    входящем в состав библиотеки.

    ДАННОЕ ПРОГРАММНОЕ ОБЕСПЕЧЕНИЕ ПРЕДОСТАВЛЯЕТСЯ «КАК ЕСТЬ», БЕЗ
    КАКИХ-ЛИБО ГАРАНТИЙ, ЯВНО ВЫРАЖЕННЫХ ИЛИ ПОДРАЗУМЕВАЕМЫХ,
    ВКЛЮЧАЯ ГАРАНТИИ ТОВАРНОЙ ПРИГОДНОСТИ, СООТВЕТСТВИЯ ПО ЕГО
    КОНКРЕТНОМУ НАЗНАЧЕНИЮ И ОТСУТСТВИЯ НАРУШЕНИЙ, НО НЕ ОГРАНИЧИВАЯСЬ 
    ИМИ. НИ В КАКОМ СЛУЧАЕ АВТОРЫ ИЛИ ПРАВООБЛАДАТЕЛИ НЕ НЕСУТ
    ОТВЕТСТВЕННОСТИ ПО КАКИМ-ЛИБО ИСКАМ, ЗА УЩЕРБ ИЛИ ПО ИНЫМ ТРЕБОВАНИЯМ,
    В ТОМ ЧИСЛЕ, ПРИ ДЕЙСТВИИ КОНТРАКТА, ДЕЛИКТЕ ИЛИ ИНОЙ СИТУАЦИИ,
    ВОЗНИКШИМ ИЗ-ЗА ИСПОЛЬЗОВАНИЯ ПРОГРАММНОГО ОБЕСПЕЧЕНИЯ ИЛИ ИНЫХ
    ДЕЙСТВИЙ С ПРОГРАММНЫМ ОБЕСПЕЧЕНИЕМ.

 **********************************************************************}

{$I uev3defines.inc}

unit rubiroButtons;

interface
uses cthreads,uev3,BaseUnix;

const
      bBACK = 14;
      bCENTER = 28;
      bUP = 103;
      bLEFT = 105;
      bRIGHT = 106;
      bDOWN = 108;

type
    TButtonSet=set of byte;
    { TEv3Buttons }

    TEv3Buttons=class(TEv3)
{
 Идентификация кнопок в функциях ожидания (регистр символов значения не имеет):
  Кнопка UP, варианты идентификации: 'up','вверх','103','bup'
  Кнопка DOWN, варианты идентификации: 'down','вниз','108','bdown'
  Кнопка LEFT, варианты идентификации: 'left','влево','105','bleft'
  Кнопка RIGHT, варианты идентификации: 'right','вправо','106','bright'
  Кнопка CENTER, варианты идентификации: 'center','центр','28','bcenter'
  Кнопка BACK, варианты идентификации: 'back','cancel','назад','отмена','bback'
  и все остальные варианты
 Идентификация состояния кнопок
  Кнопка отжата: 'нет','no','0',''
  Кнопка нажата: все остальные варианты
}
      private
       const
        KEY_MAX = $2ff;
        BUF_LEN = (KEY_MAX + 7) div 8;
       var
        fbytes:array[0..BUF_LEN-1]of byte;
        fi:cint;
      private
        function testbit(bit:byte):boolean;
        procedure testButtons();
      private
        function getBack: variant;
        function getCenter: variant;
        function getDown: variant;
        function getLeft: variant;
        function getRight: variant;
        function getUp: variant;
      protected
       class function varToButton(button:variant):integer;
       class function ParseEvent(event:string; out Button:integer; out Press:boolean):string;
       class function createEvent(Button:integer; Press:boolean):string;
       {
        ждет нажатия или отпускание кнопки.
        кнопка кодируется событием вида '14:0','назад:нет' (отпускание клавиши Back),
        '108:1','Down:NO' (нажатие клавиши Down) и т.п. Регистр символов неважен.
        timeout - максимальный таймаут
        возвращает false при выходе по таймауту
       }
       function waitEvent(event:string; timeout:dword=high(dword)):boolean; override;
      public
       constructor Create();
       destructor destroy;override;
       {
        Ждет нажатия или отпускания клавиши, в Button возвращает одну из
        констант bBACK, bCENTER, bUP, bLEFT, bRIGHT или bDOWN,
        в Press возвращает true, если было нажатие.
       }
       function wait(out Button:variant; out Press:variant):TEv3Buttons;
       {
        Ждет нажатия-отпускания набора кнопок.
        Завершение - при возникновении ВСЕХ событий из переданного массива.
        Каждая кнопка в массиве кодируется строкой вида '14:0','назад:нет'
        (отпускание клавиши Back), '108:1','Down:NO' (нажатие клавиши Down)
        и т.п. Регистр символов неважен. Процесс ожидания сопровождается
        миганием цветоиндикации
       }
       function waitAll(Buttons:array of const):TEv3Buttons;
       {
        ждать нажатия-отпускания набора кнопок
        завершение - при возникновении ЛЮБОГО события из переданного массива
        каждая кнопка в массиве кодируется строкой (см.WaitEvent)
        процессе ожидания сопровождается миганием цветоиндикации
       }
       function waitAny(Buttons:array of const):TEv3Buttons;
       // возвращает состояние кнопки BACK, true - при нажатии
       property Back:variant read getBack;
       // ожидает нажатия (1 или true) или отпускания (0 или false) кнопки BACK
       function waitBack(press:byte=1):TEv3Buttons;
       function waitBack(press:boolean):TEv3Buttons;
       // возвращает состояние кнопки CENTER, true - при нажатии
       property Center:variant read getCenter;
       // ожидает нажатия (1 или true) или отпускания (0 или false) кнопки CENTER
       function waitCenter(press:byte=1):TEv3Buttons;
       function waitCenter(press:boolean):TEv3Buttons;
       // возвращает состояние кнопки UP, true - при нажатии
       property Up:variant read getUp;
       // ожидает нажатия (1 или true) или отпускания (0 или false) кнопки UP
       function waitUp(press:byte=1):TEv3Buttons;
       function waitUp(press:boolean):TEv3Buttons;
       // возвращает состояние кнопки LEFT, true - при нажатии
       property Left:variant read getLeft;
       // ожидает нажатия (1 или true) или отпускания (0 или false) кнопки LEFT
       function waitLeft(press:byte=1):TEv3Buttons;
       function waitLeft(press:boolean):TEv3Buttons;
       // возвращает состояние кнопки RIGHT, true - при нажатии
       property Right:variant read getRight;
       // ожидает нажатия (1 или true) или отпускания (0 или false) кнопки RIGHT
       function waitRight(press:byte=1):TEv3Buttons;
       function waitRight(press:boolean):TEv3Buttons;
       // возвращает состояние кнопки DOWN, true - при нажатии
       property Down:variant read getDown;
       // ожидает нажатия (1 или true) или отпускания (0 или false) кнопки DOWN
       function waitDown(press:byte=1):TEv3Buttons;
       function waitDown(press:boolean):TEv3Buttons;
       function getButtons:TButtonSet;

  end;

var ev3buttons:TEv3Buttons;
implementation
uses
  SysUtils,Unix,uev3func,strutils,dateutils;

{ TEv3Buttons }

function TEv3Buttons.testbit(bit: byte): boolean;
begin
 if (fbytes[bit div 8] and (1 shl (bit mod 8)))<>0 then exit(true)
 else exit(false);
end;

procedure TEv3Buttons.testButtons;
begin
 _lock(9);
 try
   fpioctl(fi, EVIOCGKEY(sizeof(fbytes)), @fbytes);
 finally
   _unlock(9);
 end;
end;

function TEv3Buttons.getBack: variant;
begin
 testbuttons;
 result:=testbit(bBack);
end;

function TEv3Buttons.getCenter: variant;
begin
  testbuttons;
  result:=testbit(bCenter);
end;

function TEv3Buttons.getDown: variant;
begin
  testbuttons;
  result:=testbit(bDown);
end;

function TEv3Buttons.getLeft: variant;
begin
  testbuttons;
  result:=testbit(bLeft);
end;

function TEv3Buttons.getRight: variant;
begin
  testbuttons;
  result:=testbit(bRight);
end;

function TEv3Buttons.getUp: variant;
begin
  testbuttons;
  result:=testbit(bUp);
end;

constructor TEv3Buttons.Create;
begin
  inherited create;
  fi:=fpopen('/dev/input/by-path/platform-gpio-keys.0-event',O_RDONLY);
  if fi<0 then fi:=fpopen('/dev/input/by-path/platform-gpio_keys-event',O_RDONLY);
end;

destructor TEv3Buttons.destroy;
begin
  fpclose(fi);
  inherited destroy;
end;

__WARNOFFImplicitConv
class function TEv3Buttons.varToButton(button: variant): integer;
begin
 case UTF8LowerCase(trim(button)) of
  'up','bup','вверх','103':result:=bUP;
  'down','bdown','вниз','108':result:=bDown;
  'left','bleft','влево','105':result:=bLeft;
  'right','bright','вправо','106':result:=bRight;
  'center','bcenter','центр','28':result:=bCenter;
  'back','bback','cancel','назад','отмена','14':result:=bBack;
  else result:=bBack;
 end;
end;
__WARNON

class function TEv3Buttons.ParseEvent(event: string; out Button: integer; out
  Press: boolean):string;
var
 b:integer; p:boolean;
begin
 b:=varToButton(extractWord(1,event,[':']));
 p:=varToYes(extractWord(2,event,[':']));
 button:=b;
 press:=p;
 result:=createEvent(b,p);
end;

function TEv3Buttons.waitEvent(event: string; timeout: dword): boolean;
 var butt:cint; push:boolean;
     res:boolean;
     tm:double;
begin
 parseEvent(event,butt,push);
// writeln(butt,' ',push);
// dt:=now();
 _timerStart(tm);
 while true do begin
   testbuttons;
   res:=testbit(butt);
   //writeln(res,' ',push);
   if res=push then begin
    exit(true);
   end;
   //if MilliSecondsBetween(now(),dt)>=timeout then exit(false);
   // v0.2
   if _timer(tm)>timeout then exit(false);
   sleep(waitInterval)
 end;
end;

class function TEv3Buttons.createEvent(Button: integer; Press: boolean): string;
begin
  result:=intToStr(Button)+':'+intToStr(ord(Press));
end;

function TEv3Buttons.wait(out Button: variant; out Press:variant): TEv3Buttons;
type TButtonRec=packed record
      seconds,microseconds:dword;
      tp:word;
      code:word;
      value:dword;
     end;
  var ev3b:TButtonRec;
begin
 result:=self;
 __WARNOFFVarNotInit
 repeat
    fpread(fi,ev3b,sizeof(ev3b));
 until ev3b.tp<>0;
 __WARNON
 Button:=ev3b.code; Press:=boolean(ev3b.value);
end;

function TEv3Buttons.waitAll(Buttons: array of const): TEv3Buttons;
var o:array of TEv3=nil; e:array of string=nil;
    i:integer; R:TEv3;
begin
  result:=self;
  setlength(o,(High(Buttons)+1));
  setlength(e,(High(Buttons)+1));
  for i:=0 to high(Buttons) do begin
   o[i]:=self;
   e[i]:=arg2String(buttons[i]);
  end;
  waitEvents(o,e,evAND,R);
end;

function TEv3Buttons.waitAny(Buttons: array of const): TEv3Buttons;
var o:array of TEv3=nil; e:array of string=nil;
    i:integer; R:TEv3;
begin
  result:=self;
  setlength(o,(High(Buttons)+1));
  setlength(e,(High(Buttons)+1));
  for i:=0 to high(Buttons) do begin
   o[i]:=self;
   e[i]:=arg2String(buttons[i]);
  end;
  waitEvents(o,e,evOr,R);
end;

function TEv3Buttons.waitBack(press: byte): TEv3Buttons;
begin
  result:=self;
  waitAny(['back:'+inttostr(press)]);
end;

function TEv3Buttons.waitBack(press: boolean): TEv3Buttons;
begin
 result:=waitBack(ord(press));
end;

function TEv3Buttons.waitCenter(press: byte): TEv3Buttons;
begin
 result:=self;
 waitAny(['center:'+inttostr(press)]);
end;

function TEv3Buttons.waitCenter(press: boolean): TEv3Buttons;
begin
 result:=waitCenter(ord(press));
end;

function TEv3Buttons.waitUp(press: byte): TEv3Buttons;
begin
 result:=self;
 waitAny(['up:'+inttostr(press)]);
end;

function TEv3Buttons.waitUp(press: boolean): TEv3Buttons;
begin
 result:=waitUp(ord(press));
end;

function TEv3Buttons.waitLeft(press: byte): TEv3Buttons;
begin
 result:=self;
 waitAny(['left:'+inttostr(press)]);
end;

function TEv3Buttons.waitLeft(press: boolean): TEv3Buttons;
begin
 result:=waitLeft(ord(press));
end;

function TEv3Buttons.waitRight(press: byte): TEv3Buttons;
begin
 result:=self;
 waitAny(['right:'+inttostr(press)]);
end;

function TEv3Buttons.waitRight(press: boolean): TEv3Buttons;
begin
 result:=waitRight(ord(press));
end;

function TEv3Buttons.waitDown(press: byte): TEv3Buttons;
begin
 result:=self;
 waitAny(['down:'+inttostr(press)]);
end;

function TEv3Buttons.waitDown(press: boolean): TEv3Buttons;
begin
 result:=waitDown(ord(press));
end;

function TEv3Buttons.getButtons: TButtonSet;
begin
  testbuttons;
  result:=[];
  if testbit(bBack) then include(result,bBack);
  if testbit(bCENTER) then include(result,bCENTER);
  if testbit(bUP) then include(result,bUP);
  if testbit(bDown) then include(result,bDown);
  if testbit(bLeft) then include(result,bLeft);
  if testbit(bRight) then include(result,bRight);
end;

procedure init;
  begin
    fpsystem('stty -echoctl 2>/dev/null');
    ev3buttons:=Tev3buttons.Create;
  end;

procedure done; begin ev3buttons.free;fpsystem('stty echoctl 2>/dev/null'); end;

initialization
   PushInit(@init);
   PushDone(@Done);
end.

